# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution    
#    Copyright (C) 2004-2008 Tiny SPRL (<http://tiny.be>). All Rights Reserved
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import fields, osv
from tools import config
import time
from tools.translate import _

class account_payment_term(osv.osv):
    _name = "account.payment.term"
    _inherit = "account.payment.term"
    _columns = {
        'cash_discount_ids': fields.one2many('account.cash.discount', 'payment_id', 'Cash Discounts'),
    }
account_payment_term()

class account_cash_discount(osv.osv):
    _name = "account.cash.discount"
    _description = "Cash Discount" #A reduction in the price  if payment is made within a stipulated period.
    _order ="delay"
    _columns = {
        'name': fields.char('Name', size=32),
        'delay': fields.integer('Number of Days', required=True),
        'discount': fields.float('Discount (%)', digits=(16,6),required=True),
        'discount_account_id': fields.many2one('account.account', 'Discount Account',required=True),
        'payment_id': fields.many2one('account.payment.term','Associated Payment Term'),
        'credit_account_id': fields.many2one('account.account', 'Credit Account'),
        'debit_account_id': fields.many2one('account.account', 'Debit Account'),
#        'payment_term_ids': fields.one2many('account.payment.term.line', 'cash_account_discount_id', 'Payment Term Lines'),
#        'day_tolerance': fields.integer('Days Tolerance', digits=(16,6),required=True),
        'sup_discount_account_id': fields.many2one('account.account', 'Supplier Discount Account',required=True),
    }
account_cash_discount()

class account_payment_term_line(osv.osv):
    _inherit = "account.payment.term.line"
    _columns = {
#        'cash_account_discount_id': fields.many2one('account.cash.discount', 'Discount Lines',),
        'day_tolerance': fields.integer('Days Tolerance', digits=(16,6),required=True),
        'payment_id': fields.many2one('account.payment.term', 'Payment Term', required=False, select=True),
        'compl_cash_discount':fields.boolean('Use Complete Cash Discount'),
    }
account_payment_term_line()

class account_invoice(osv.osv):
    _inherit = "account.invoice"
    
    def _get_amount(self, cr, uid, ids, resudial_amonut, payment_term, context=None):
        """
        This function return the Amount to paid  according to the payment term cash discount payment term lines
        """
        if context is None:
           context = {}
        tax_obj = self.pool.get('account.tax')
        invoice = self.browse(cr, uid, ids[0], context=context)
        
        if invoice.date_invoice:
            date1 = invoice.date_invoice
        else:
            date1 =  time.strftime('%Y-%m-%d')
                
        if 'date_p' in context and context['date_p']:    
            date2 = context['date_p']
        else:
            date2 =  time.strftime('%Y-%m-%d')   

        if date1 >= date2:
            from_dt = time.mktime(time.strptime(date1,'%Y-%m-%d'))
            to_dt = time.mktime(time.strptime(date2,'%Y-%m-%d'))
            diff_day = (from_dt-to_dt)/(3600*24)
        else:
            from_dt = time.mktime(time.strptime(date2,'%Y-%m-%d'))
            to_dt = time.mktime(time.strptime(date1,'%Y-%m-%d'))
            diff_day = (from_dt-to_dt)/(3600*24)
        
        discount = 0.0
        amount = 0.0
        tol_days = 0.0
        payment_line_found = False
          
        if payment_term:
            payment_term_lines = self.pool.get('account.payment.term').browse(cr, uid, payment_term, context=context)
            if payment_term_lines.line_ids:
                    res = 0.0
                    dis = 0
                    for payment_line in payment_term_lines.line_ids:
                        if diff_day >= dis and  diff_day <= payment_line.days:
                            if resudial_amonut:
                                payment_line_found = True
                                tol_days = payment_line.day_tolerance
                                if payment_line.value == 'procent':
                                    res = resudial_amonut * payment_line.value_amount
                                elif payment_line.value == 'fixed':
                                    res = payment_line.value_amount
                                else:
                                    res = resudial_amonut
                            amount = res
                        dis = payment_line.days
                        
        return {'amount':amount,'tol_days':tol_days,'found':payment_line_found}
     
    def _get_payment(self, cr, uid, ids, resudial_amonut, payment_term, context=None):
        """
        This function return the Discount according to the payment term cash discount term
        """
        if context is None:
           context = {}
           
        tax_obj = self.pool.get('account.tax')
        invoice = self.browse(cr, uid, ids[0], context=context)
        
        if invoice.date_invoice:
            date1 = invoice.date_invoice
        else:
            date1 =  time.strftime('%Y-%m-%d')
                
        if 'date_p' in context and context['date_p']:    
            date2 = context['date_p']
        else:
            date2 =  time.strftime('%Y-%m-%d')   

        if date1 >= date2:
            from_dt = time.mktime(time.strptime(date1,'%Y-%m-%d'))
            to_dt = time.mktime(time.strptime(date2,'%Y-%m-%d'))
            diff_day = (from_dt-to_dt)/(3600*24)
        else:
            from_dt = time.mktime(time.strptime(date2,'%Y-%m-%d'))
            to_dt = time.mktime(time.strptime(date1,'%Y-%m-%d'))
            diff_day = (from_dt-to_dt)/(3600*24)
        
        tolerance = self._get_amount(cr, uid, ids, resudial_amonut, payment_term, context)
        tolerance_days = tolerance['tol_days']
        
        discount = 0.0
        if payment_term:
            payment_term_lines = self.pool.get('account.payment.term').browse(cr, uid, payment_term, context=context)
            if payment_term_lines.cash_discount_ids:
                    res = 0.0
                    dis = 0
                    for discount_line in payment_term_lines.cash_discount_ids:
                        if diff_day >= dis and  diff_day <= discount_line.delay + tolerance_days:
                            if resudial_amonut:
                                res = resudial_amonut * discount_line.discount
                            discount = res
                        dis = discount_line.delay + tolerance_days
        return discount
    
    def _get_account(self, cr, uid, ids, resudial_amonut, payment_term, context=None):
        """
        This function return the Account according to the payment term cash discount term
        """
        if context is None:
           context = {}
        
        account_id = False
           
        tax_obj = self.pool.get('account.tax')
        invoice = self.browse(cr, uid, ids[0], context=context)
        
        if invoice.date_invoice:
            date1 = invoice.date_invoice
        else: 
            date1 =  time.strftime('%Y-%m-%d')
                
        if 'date_p' in context and context['date_p']:    
            date2 = context['date_p']
        else:
            date2 =  time.strftime('%Y-%m-%d')   
        
        if date1 > date2:
            from_dt = time.mktime(time.strptime(date1,'%Y-%m-%d'))
            to_dt = time.mktime(time.strptime(date2,'%Y-%m-%d'))
            diff_day = (from_dt-to_dt)/(3600*24)
        else:
            from_dt = time.mktime(time.strptime(date2,'%Y-%m-%d'))
            to_dt = time.mktime(time.strptime(date1,'%Y-%m-%d'))
            diff_day = (from_dt-to_dt)/(3600*24)    
        
        tolerance = self._get_amount(cr, uid, ids, resudial_amonut, payment_term, context)
        tolerance_days = tolerance['tol_days']
        
        if payment_term:
            payment_term_lines = self.pool.get('account.payment.term').browse(cr, uid, payment_term, context=context)
            if payment_term_lines.cash_discount_ids:
                    res = 0.0
                    dis = 0.0
                    for discount_line in payment_term_lines.cash_discount_ids:
                        if diff_day >= dis and  diff_day <= discount_line.delay + tolerance_days:
                            if invoice.type in ('in_invoice', 'in_refund'):
                                account_id = discount_line.sup_discount_account_id.id
                            else:    
                                account_id = discount_line.discount_account_id.id
                        dis = discount_line.delay + tolerance_days
                    return account_id 
            
    def pay_and_reconcile(self, cr, uid, ids, pay_amount, pay_account_id, period_id, pay_journal_id, writeoff_acc_id, writeoff_period_id, writeoff_journal_id, context=None, name=''):
        if context is None:
            context = {}
        #TODO check if we can use different period for payment and the writeoff line
        assert len(ids)==1, "Can only pay one invoice at a time"
        invoice = self.browse(cr, uid, ids[0])
        
        tax_obj = self.pool.get('account.tax')
        invoice_tax_obj = self.pool.get("account.invoice.tax")
        move_line_obj = self.pool.get('account.move.line')
        move_obj = self.pool.get('account.move')
        
        amount_discount = 0.0    
        if invoice.payment_term:
            # Return the discount on for the payment term
            amount_discount = self._get_payment(cr, uid, ids, pay_amount, invoice.payment_term.id, context=context)

        
        src_account_id = invoice.account_id.id
        # Take the seq as name for move
        types = {'out_invoice': -1, 'in_invoice': 1, 'out_refund': 1, 'in_refund': -1}
        direction = types[invoice.type]
        #take the choosen date
        if 'date_p' in context and context['date_p']:
            date=context['date_p']
        else:
            date=time.strftime('%Y-%m-%d')
            
        # Take the amount in currency and the currency of the payment
        if 'amount_currency' in context and context['amount_currency'] and 'currency_id' in context and context['currency_id']:
            amount_currency = context['amount_currency']
            currency_id = context['currency_id']
        else:
            amount_currency = False
            currency_id = False
        
        if invoice.type in ('in_invoice', 'in_refund'):
            ref = invoice.reference
        else:
            ref = self._convert_ref(cr, uid, invoice.number)
        
        discount_amount = 0.0
        tax_total_amount = 0.0
        
        if 'cash_amount' in context and context['cash_amount']:
            discount_amount = context['cash_amount']
            
        # Pay attention to the sign for both debit/credit AND amount_currency
        l1 = {
            'debit': direction * (pay_amount)>0 and direction * (pay_amount),
            'credit': direction * (pay_amount)<0 and - direction * (pay_amount),
            'account_id': src_account_id,
            'partner_id': invoice.partner_id.id,
            'ref':ref,
            'date': date,
            'currency_id':currency_id,
            'amount_currency':amount_currency and direction * amount_currency or 0.0,
        }
        
        ### if we change the amount paid and in the disocunt  then it used the changes discount
            
        if amount_discount != discount_amount and discount_amount>0.0:
            amount_discount = discount_amount
            
        l2 = {
            'debit': direction * (pay_amount-amount_discount)<0 and - direction * (pay_amount-amount_discount),
            'credit': direction * (pay_amount-amount_discount)>0 and direction * (pay_amount-amount_discount),
            'account_id': pay_account_id,
            'partner_id': invoice.partner_id.id,
            'ref':ref,
            'date': date,
            'currency_id':currency_id,
            'amount_currency':amount_currency and - direction * amount_currency or 0.0,
        }

        if not name:
            name = invoice.invoice_line and invoice.invoice_line[0].name or invoice.number
            
        l1['name'] = name
        l2['name'] = name
        lines = [(0, 0, l1), (0, 0, l2)]
        
        ### When we make payment by pay invoice wizard and if found the taxes move then directly take the entry from the wizard  
        if 'tax_move_ids' in context and context['tax_move_ids']:
            move_line = context['tax_move_ids']
            for move_line_id in move_line:
                move_line_data = move_line_obj.browse(cr, uid,move_line_id)
                l3 = {
                        'debit': move_line_data.debit,
                        'credit': move_line_data.credit,
                        'account_id': move_line_data.account_id.id,
                        'partner_id': move_line_data.partner_id.id,
                        'ref':move_line_data.ref,
                        'date': move_line_data.date,
                        'currency_id':currency_id,
                        'amount_currency':amount_currency and direction * amount_currency or 0.0,
                        'name':move_line_data.name,
                        'tax_code_id':move_line_data.tax_code_id.id,
                        'tax_amount':move_line_data.tax_amount,
                        }
                lines.append((0, 0, l3))
                move_line_obj.unlink(cr, uid,[move_line_id])
                move_obj.unlink(cr, uid,[move_line_data.move_id.id])
        else:
            if amount_discount>0.0:         
                for line in invoice.invoice_line:
                    if line.invoice_line_tax_id:
                        for tax in tax_obj.compute(cr, uid, line.invoice_line_tax_id, amount_discount, line.quantity, invoice.address_invoice_id.id, line.product_id, invoice.partner_id):
                            tax_amount = tax['amount']
                            tax_total_amount += tax['amount']
                            tax_invoice_id = invoice_tax_obj.search(cr,uid,[('invoice_id','=',invoice.id),('name','=',tax['name'])])
                            for tax_id in tax_invoice_id:
                                tax_value = invoice_tax_obj.browse(cr, uid, tax_id)
                                tax_account_id = tax_value.account_id.id
                            l3 = {
                                'debit': direction * tax_amount<0 and - direction * tax_amount,
                                'credit': direction * tax_amount>0 and  direction * tax_amount,
                                'account_id': tax_account_id,
                                'partner_id': invoice.partner_id.id,
                                'ref':ref,
                                'date': date,
                                'currency_id':currency_id,
                                'amount_currency':amount_currency and direction * amount_currency or 0.0,
                                'tax_code_id':tax_value.tax_code_id.id,
                                'tax_amount':tax_amount,
                            }
                            l3['name'] = name
                            lines.append((0, 0, l3))
                            
        ### When we make payment by pay invoice wizard and if found the discount moves in wizard discount move then directly take the entry from the wizard
        if 'discount_move_ids' in context and context['discount_move_ids']: 
            move_line = context['discount_move_ids']
            for move_line_id in move_line:
                move_line_data = move_line_obj.browse(cr, uid,move_line_id)
                l4 = {
                        'debit': move_line_data.debit,
                        'credit':move_line_data.credit,
                        'account_id': move_line_data.account_id.id,
                        'partner_id': move_line_data.partner_id.id,
                        'ref':move_line_data.ref,
                        'date': move_line_data.date,
                        'currency_id':currency_id,
                        'amount_currency':amount_currency and direction * amount_currency or 0.0,
                        'name':move_line_data.name,
                        'tax_code_id':move_line_data.tax_code_id.id,
                        'tax_amount':move_line_data.tax_amount,
                      }
                lines.append((0, 0, l4))
                move_line_obj.unlink(cr, uid,[move_line_id])
                move_obj.unlink(cr, uid,[move_line_data.move_id.id])
        else:
            if amount_discount>0:
                if 'account_id' in context and context['account_id']:
                    account_id = context['account_id']  
                if tax_total_amount>0:
                    amount_discount = (amount_discount - tax_total_amount)
                else:
                    amount_discount = amount_discount
                l4 = {
                    'debit': direction * amount_discount<0 and - direction * amount_discount,
                    'credit': direction * amount_discount>0 and  direction * amount_discount,
                    'account_id': account_id,
                    'partner_id': invoice.partner_id.id,
                    'ref':ref,
                    'date': date,
                    'currency_id':currency_id,
                    'amount_currency':amount_currency and direction * amount_currency or 0.0,
                }
                l4['name'] = name
                lines.append((0, 0, l4))
        
        move = {'ref': ref, 'line_id': lines, 'journal_id': pay_journal_id, 'period_id': period_id, 'date': date}
        move_id = move_obj.create(cr, uid, move, context=context)
        
        line_ids = []
        total = 0.0
        line = self.pool.get('account.move.line')
        cr.execute('select id from account_move_line where move_id in ('+str(move_id)+','+str(invoice.move_id.id)+')')
        
        lines = line.browse(cr, uid, map(lambda x: x[0], cr.fetchall()) )
        for l in lines+invoice.payment_ids:
            if l.account_id.id==src_account_id:
                line_ids.append(l.id)
                total += (l.debit or 0.0) - (l.credit or 0.0)
        if (not round(total,int(config['price_accuracy']))) or writeoff_acc_id:
            self.pool.get('account.move.line').reconcile(cr, uid, line_ids, 'manual', writeoff_acc_id, writeoff_period_id, writeoff_journal_id, context)
        else:
            self.pool.get('account.move.line').reconcile_partial(cr, uid, line_ids, 'manual', context)

        # Update the stored value (fields.function), so we write to trigger recompute
        self.pool.get('account.invoice').write(cr, uid, ids, {}, context=context)
        return True
    
account_invoice()

class account_move_line(osv.osv):
    _name = 'account.move.line'
    _inherit = 'account.move.line'
    
    def _invoice(self, cr, uid, ids, name, arg, context=None):
        res = {}
        if ids:
            return super(account_move_line, self)._invoice(cr, uid, ids, name, arg, context)
        else:
            return res
        
    def _invoice_search(self, cr, uid, obj, name, args, context={}):
        return super(account_move_line, self)._invoice_search(cr, uid, obj, name, args, context=context)
        
    _columns = {
       'invoice': fields.function(_invoice, method=True, string='Invoice',
          type='many2one', relation='account.invoice', fnct_search=_invoice_search),
    }
account_move_line()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:

