# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import time
from lxml import etree
from osv import fields, osv
from tools.translate import _

class account_invoice_pay_writeoff(osv.osv_memory):
    """
    Opens the write off amount pay form.
    """
    _name = "account.invoice.pay.writeoff"
    _description = "Pay Invoice  "
    _columns = {
        'writeoff_acc_id': fields.many2one('account.account', 'Write-Off account', required=True),
        'writeoff_journal_id': fields.many2one('account.journal', 'Write-Off journal', required=True),
        'comment': fields.char('Comment', size=64, required=True),
        'analytic_id': fields.many2one('account.analytic.account','Analytic Account'),
        }
    _defaults = {
        'comment':lambda *a :'Write-Off',
        }

    def pay_and_reconcile_writeoff(self, cr, uid, ids, context=None):
        pay_invoice = self.pool.get('account.invoice.pay')
        data =  self.read(cr, uid, ids,context=context)[0]
        context.update({'write_off':data})
        self.pool.get('account.invoice.pay').pay_and_reconcile(cr, uid, ids, context=context)
        return {}

account_invoice_pay_writeoff()

class account_invoice_pay(osv.osv_memory):
    """
    Generate pay invoice wizard, user can make partial or full payment for invoice.
    """
    _name = "account.invoice.pay"
    _description = "Pay Invoice  "
    _columns = {
        'amount': fields.float('Amount paid', required=True),
        'name': fields.char('Entry Name', size=64, required=True),
        'date': fields.date('Date payment', required=True),
        'cash_residual_amount': fields.float('Residual Amount'),
        'journal_id': fields.many2one('account.journal', 'Journal/Payment Mode', required=True),
        'period_id': fields.many2one('account.period', 'Period', required=True),
        'cash_amount':fields.float('Cash Discount Amount',),
        'account_id': fields.many2one('account.account', 'Customers Cash Discount Account',),
        'cal_method_selection': fields.selection([('method_gross_methodology','Calculate Discount with Net Gross Methodology'),
                                            ('method_net_methodology','Calculate Discount with Net Methodology'),],
                                              'Method To Calculation'),
        'discount_move_ids': fields.many2many('account.move.line', 'account_discount_move_rel', 'discount_account_id', 'discount_move_id', 'Account Discount Moves'),
        'tax_move_ids': fields.many2many('account.move.line', 'account_tax_move_rel', 'tax_account_id', 'tax_move_id', 'Account Taxes Moves'),
        }
    
    def _get_period(self, cr, uid, context=None):
        ids = self.pool.get('account.period').find(cr, uid, context=context)
        period_id = False
        if len(ids):
            period_id = ids[0]
        return period_id

    def _get_amount(self, cr, uid, context=None):
        obj_inv = self.pool.get('account.invoice')
        tax_obj = self.pool.get('account.tax')
        invoice =  obj_inv.browse(cr, uid, context['id'], context=context)
        total_tax_amount = 0.0 
        total_amount = 0.0
        amount = obj_inv._get_amount(cr, uid, [context['id']] ,invoice.residual, invoice.payment_term.id, context=context)
        context.update({'found':amount['found']})
        amount = amount['amount']
        discount = obj_inv._get_payment(cr, uid, [context['id']] ,amount, invoice.payment_term.id, context=context)
        residual_amount = amount - discount
        return residual_amount
    
    def _get_discount(self, cr, uid, context=None):
        obj_inv = self.pool.get('account.invoice')
        tax_obj = self.pool.get('account.tax')
        invoice =  obj_inv.browse(cr, uid, context['id'], context=context)
        amount = obj_inv._get_amount(cr, uid, [context['id']] ,invoice.residual, invoice.payment_term.id, context=context)
        amount = amount['amount']
        discount = obj_inv._get_payment(cr, uid, [context['id']] , amount, invoice.payment_term.id, context=context)
        return discount
    
    def _get_account(self, cr, uid, context=None):
        """
        This function return the Account according to the payment term cash term and map the account with fiscal position
        """
        obj_inv = self.pool.get('account.invoice')
        invoice =  obj_inv.browse(cr, uid, context['id'], context=context)
        account = obj_inv._get_account(cr, uid, [context['id']] , invoice.residual, invoice.payment_term.id, context=context)
        fpos = invoice.fiscal_position or False
        if account:
            account = self.pool.get('account.fiscal.position').map_account(cr, uid, fpos, account)
        return account
    
    _defaults = {
        'date': lambda *a: time.strftime('%Y-%m-%d'),
        'period_id': _get_period,
        'amount': _get_amount,
        'cash_amount':_get_discount,
        'account_id':_get_account,
        'cash_residual_amount': lambda *a:0.0,
        'cal_method_selection':lambda *a :'method_gross_methodology',
        }
    
    def on_change_ammount(self, cr, uid, ids, amount, context=None):
        """
        Function return the Discount according to the Amount paid and Payment Term Cash Discount
        """
        res = {}
        obj_inv = self.pool.get('account.invoice')
        invoice =  obj_inv.browse(cr, uid, context['id'], context=context)
        cal_amount = obj_inv._get_amount(cr, uid, [context['id']] ,invoice.residual, invoice.payment_term.id, context=context)
        
        old_cal_amount = cal_amount['amount']
        discount = obj_inv._get_payment(cr, uid, [context['id']] , old_cal_amount, invoice.payment_term.id, context=context)
        old_amount = old_cal_amount 
        diff_amount = round(old_amount -  (amount + discount),2)
        
        return {'value' : {'cash_residual_amount':diff_amount}}
    
    def on_change_cash_discount_amount(self, cr, uid, ids, discount_amount, amount, context=None):
        ###  Return discount amount
        res = {}
        obj_inv = self.pool.get('account.invoice')
        invoice =  obj_inv.browse(cr, uid, context['id'], context=context)
        cal_amount = obj_inv._get_amount(cr, uid, [context['id']] ,invoice.residual, invoice.payment_term.id, context=context)
        old_cal_amount = cal_amount['amount']
        discount = obj_inv._get_payment(cr, uid, [context['id']] , amount, invoice.payment_term.id, context=context)
        old_amount = old_cal_amount

        diff_amount = 0.0
        diff_amount = old_amount -  (amount + discount_amount)
        
        return {'value' : {'cash_amount':discount_amount,'cash_residual_amount':diff_amount}}
    
    def _calculation(self, cr, uid, ids, context=None):

        invoice_obj = self.pool.get('account.invoice')
        data =  self.read(cr, uid, ids,context=context)[0]
        invoice = self.pool.get('account.invoice').browse(cr, uid, context['id'], context=context)
        tax_obj = self.pool.get('account.tax')
        invoice_tax_obj = self.pool.get("account.invoice.tax")
        move_obj = self.pool.get('account.move')
        move_line_obj = self.pool.get('account.move.line')

        ############## Delete another lines of taxes and discount
        if data['tax_move_ids']:
            move_line = data['tax_move_ids']
            for move_line_id in move_line:
                move_id = move_line_obj.browse(cr, uid,move_line_id).move_id.id
                move_line_obj.unlink(cr, uid,[move_line_id])
                move_obj.unlink(cr, uid,[move_id])
                
        if data['discount_move_ids']: 
            discount_move_line = data['discount_move_ids']
            for move_line_id in discount_move_line:
                move_id = move_line_obj.browse(cr, uid,move_line_id).move_id.id
                move_line_obj.unlink(cr, uid,[move_line_id])
                move_obj.unlink(cr, uid,[move_id])
            
        ######## to get ref
        if invoice.type in ('in_invoice', 'in_refund'):
            ref = invoice.reference
        else:
            ref = invoice_obj._convert_ref(cr, uid, invoice.number)        
        
        #### to get direction
        types = {'out_invoice': -1, 'in_invoice': 1, 'out_refund': 1, 'in_refund': -1}
        direction = types[invoice.type]
        date=time.strftime('%Y-%m-%d')
        
        #### to get name
        name = invoice.invoice_line and invoice.invoice_line[0].name or invoice.number
        ##### Entry in the discount account moves and entry in the tax account moves

        if data.get('cash_amount',0.0)>0.0:
            tax_total_amount = 0.0
            amount = 0.0
            discount = 0.0
            line_ids = []
            discount_line_ids = []

            discount_account_id = False
            tax_code_id = False
            
            for line in invoice.invoice_line:
                tax_amount = 0.0
                line_tax_amount = 0.0
                
                #####get the ratio of the line in the total invoice amount 
                for tax in tax_obj.compute(cr, uid, line.invoice_line_tax_id, line.price_subtotal, line.quantity, invoice.address_invoice_id.id, line.product_id, invoice.partner_id):    
                    line_tax_amount = tax['amount']
                line_total = line.price_subtotal + line_tax_amount
                invoice_res_amount = invoice.amount_total
                line_ratio = line_total / invoice_res_amount
                ###########################
                
                if line.invoice_line_tax_id:
                    for tax in tax_obj.compute(cr, uid, line.invoice_line_tax_id, data['cash_amount'], line.quantity, invoice.address_invoice_id.id, line.product_id, invoice.partner_id):
                        tax_amount = tax['amount']
                        tax_invoice_id = invoice_tax_obj.search(cr,uid,[('invoice_id','=',invoice.id),('name','=',tax['name'])])
                        for tax_id in tax_invoice_id:
                            tax_value = invoice_tax_obj.browse(cr, uid, tax_id)
                            tax_account_id = tax_value.account_id.id
                            tax_code_id = tax_value.tax_code_id.id
                            
                        l3 = {
                            'debit': direction * tax_amount<0 and - direction * tax_amount,
                            'credit': direction * tax_amount>0 and  direction * tax_amount,
                            'account_id': tax_account_id,
                            'partner_id': invoice.partner_id.id,
                            'ref':ref,
                            'date': date,
                            'journal_id':data.get('journal_id',False) ,
                            'period_id':data.get('period_id',False),
                            'name':name,
                            'tax_code_id':tax_value.tax_code_id.id,
                            'tax_amount':-tax_amount,
                        }
                        line_id = move_line_obj.create(cr, uid, l3, context=context)
                        line_ids.append(line_id)
                
                ### if calculation on the Net Methology then in discount field it will take product income account or product category income account 
                
                if data['cal_method_selection'] == 'method_net_methodology':
                    if line.product_id.property_account_income:
                        discount_account_id = line.product_id.property_account_income.id
                    else:
                        discount_account_id = line.product_id.categ_id.property_account_income_categ.id
                else:
                     discount_account_id = data['account_id']
                
                discount = data['cash_amount'] * line_ratio
                l4 = {
                    'debit': direction * (discount - tax_amount)<0 and - direction * (discount - tax_amount),
                    'credit': direction * (discount - tax_amount)>0 and  direction * (discount - tax_amount),
                    'account_id': discount_account_id,
                    'partner_id': invoice.partner_id.id,
                    'ref':ref,
                    'date': date,
                    'name':name,
                    'journal_id':data.get('journal_id',False),
                    'period_id':data.get('period_id',False),
                    'tax_code_id':tax_code_id,
                    'tax_amount':-tax_amount,
                    }
                
                discount_line_id = move_line_obj.create(cr, uid, l4, context=context)
                discount_line_ids.append(discount_line_id)
                        
            self.write(cr, uid, ids, {'tax_move_ids':[(6,0,line_ids)]}, context)
            self.write(cr, uid, ids, {'discount_move_ids':[(6,0,discount_line_ids)]}, context)
            
        return True
    
    def default_get(self, cr, uid, fields, context=None):
      res = super(account_invoice_pay, self).default_get(cr, uid, fields, context=context)
      invoice = self.pool.get('account.invoice').browse(cr, uid, context['id'], context=context)
      if invoice.state in ['draft', 'proforma2', 'cancel','paid']:
          raise osv.except_osv(_('Error !'), _('Can not pay draft/proforma/cancel/Done invoice.'))
      return res
     
    def _message(self, cr, uid, ids, context=None):
        mod_obj = self.pool.get('ir.model.data')
        data =  self.read(cr, uid, ids,context=context)[0]
        if data.get('discount_move_ids',False):
            context.update({'discount_move_ids':data['discount_move_ids']})
        
        if data.get('id',False):
            context.update({'t_ids':ids})
            
        if data.get('tax_move_ids',False):
            context.update({'tax_move_ids':data['tax_move_ids']})

        if data.get('amount',False):
            context.update({'amount':data['amount']})
        
        if data.get('cash_amount',False):
            context.update({'cash_amount':data['cash_amount']})
                
        model_data_ids = mod_obj.search(cr, uid,[('model','=','ir.ui.view'),('name','=','view_account_message')], context=context)
        resource_id = mod_obj.read(cr, uid, model_data_ids, fields=['res_id'], context=context)[0]['res_id']
        
        return {
            'name': _('Message'),
            'context': context,
            'view_type': 'form',
            'view_mode': 'form',
            'res_model': 'account.message',
            'views': [(resource_id,'form')],
            'type': 'ir.actions.act_window',
            'target': 'new',
        } 
      
    def wo_check(self, cr, uid, ids, context=None):
        cur_obj = self.pool.get('res.currency')
        mod_obj = self.pool.get('ir.model.data')
        obj_inv = self.pool.get('account.invoice')
        
        if context is None:
            context = {}
        data =  self.read(cr, uid, ids,context=context)[0]
        invoice = self.pool.get('account.invoice').browse(cr, uid, context['id'], context)
        journal = self.pool.get('account.journal').browse(cr, uid, data['journal_id'], context)
       # Here we need that:
        #    The invoice total amount in company's currency <> paid amount in company currency
        #    (according to the correct day rate, invoicing rate and payment rate are may be different)
        #    => Ask to a write-off of the difference. This could happen even if both amount are equal,
        #    because if the currency rate
        # Get the amount in company currency for the invoice (according to move lines)
        inv_amount_company_currency = 0
        for aml in invoice.move_id.line_id:
            if aml.account_id.id == invoice.account_id.id or aml.account_id.type in ('receivable', 'payable'):
                inv_amount_company_currency += aml.debit
                inv_amount_company_currency -= aml.credit
        inv_amount_company_currency = abs(inv_amount_company_currency)
        # Get the current amount paid in company currency
        if journal.currency and invoice.company_id.currency_id.id<>journal.currency.id:
            ctx = {'date':data['date']}
            amount_paid = cur_obj.compute(cr, uid, journal.currency.id, invoice.company_id.currency_id.id, data['amount'], round=True, context=ctx)
        else:
            amount_paid = data['amount'] + data['cash_amount']
            
        # Get the old payment if there are some
        if invoice.payment_ids:
            debit=credit=0.0
            for payment in invoice.payment_ids:
                debit+=payment.debit
                credit+=payment.credit
            amount_paid+=abs(debit-credit)

        # Test if there is a difference according to currency rouding setting
#        amount = obj_inv._get_amount(cr, uid, [context['id']] ,invoice.residual, invoice.payment_term.id, context=context)
#
#        found_h = amount['found']
#
#        if found_h:

        ####### Here when we do make full payment then to get the previous Discount and Tax Move added in context
        if context.get('discount_move_ids',False):
            context.update({'discount_move_ids':context['discount_move_ids']})
        if context.get('tax_move_ids',False):
            context.update({'tax_move_ids':context['tax_move_ids']})
        if context.get('amount',False):
            context.update({'amount':context['amount']})
        if context.get('cash_amount',False):
            context.update({'cash_amount':context['cash_amount']})
        if context.get('t_ids',False):
            ids = context['t_ids']    
#            return self.pay_and_reconcile(cr, uid, ids, context=context)
#        else:
        
        if self.pool.get('res.currency').is_zero(cr, uid, invoice.company_id.currency_id,
                (amount_paid - inv_amount_company_currency)) or data['cash_residual_amount'] == 0.0:
            return self.pay_and_reconcile(cr, uid, ids, context=context)
        
        else:
            model_data_ids = mod_obj.search(cr, uid,[('model','=','ir.ui.view'),('name','=','view_account_invoice_pay_writeoff')], context=context)
            resource_id = mod_obj.read(cr, uid, model_data_ids, fields=['res_id'], context=context)[0]['res_id']
            return {
                'name': _('Information addendum'),
                'context': context,
                'view_type': 'form',
                'view_mode': 'form',
                'res_model': 'account.invoice.pay.writeoff',
                'views': [(resource_id,'form')],
                'type': 'ir.actions.act_window',
                'target': 'new',
            }

    def pay_and_reconcile(self, cr, uid, ids, context=None):
        cur_obj = self.pool.get('res.currency')
        if context is None:
            context = {}

        data =  self.read(cr, uid, ids,context=context)[0]
        writeoff_account_id = False
        writeoff_journal_id = False
        comment = False
        
        if 'write_off' in context and context['write_off'] :
            writeoff_account_id = context['write_off']['writeoff_acc_id']
            writeoff_journal_id = context['write_off']['writeoff_journal_id']
            comment = context['write_off']['comment']
        
        if context.get('amount',False):
            amount = context['amount'] +  context.get('cash_amount',0.0)
        else:
            amount = data['amount'] + data['cash_amount']
        
        invoice = self.pool.get('account.invoice').browse(cr, uid, context['id'], context=context)
        journal = self.pool.get('account.journal').browse(cr, uid, data['journal_id'], context=context)
        # Compute the amount in company's currency, with the journal currency (which is equal to payment currency)
        # when it is needed :  If payment currency (according to selected journal.currency) is <> from company currency
        if journal.currency and invoice.company_id.currency_id.id<>journal.currency.id:
            ctx = {'date':data['date']}
            amount = cur_obj.compute(cr, uid, journal.currency.id, invoice.company_id.currency_id.id, amount, context=ctx)
            currency_id = journal.currency.id
            # Put the paid amount in currency, and the currency, in the context if currency is different from company's currency
            if context.get('amount',False):
                context.update({'amount_currency':context['amount'] +  context.get('cash_amount',0.0),'currency_id':currency_id})
            else:    
                context.update({'amount_currency':data['amount'] + data['cash_amount'],'currency_id':currency_id})

        if invoice.company_id.currency_id.id<>invoice.currency_id.id:
            ctx = {'date':data['date']}
            amount = cur_obj.compute(cr, uid, invoice.currency_id.id, invoice.company_id.currency_id.id, amount, context=ctx)
            currency_id = invoice.currency_id.id
            # Put the paid amount in currency, and the currency, in the context if currency is different from company's currency
            if context.get('amount',False):
                context.update({'amount_currency':context['amount'] +  context.get('cash_amount',0.0),'currency_id':currency_id})
            else:    
                context.update({'amount_currency':data['amount'] + data['cash_amount'],'currency_id':currency_id})

        context.update({'account_id':data['account_id'],'date_p':data['date'],'comment':comment or False})        

        if context.get('discount_move_ids',False):
            context.update({'discount_move_ids':context['discount_move_ids']})
        else:
            context.update({'discount_move_ids':data['discount_move_ids']})    

        if context.get('tax_move_ids',False):
            context.update({'tax_move_ids':context['tax_move_ids']})
        else:
            context.update({'tax_move_ids':data['tax_move_ids']})
                
        if context.get('cash_amount',False):
            cash_amount = context['cash_amount']
            context.update({'cash_amount':context['cash_amount']})        
        else:
            cash_amount = data['cash_amount']
            context.update({'cash_amount':data['cash_amount']})
        
        if journal.currency and invoice.company_id.currency_id.id<>journal.currency.id:
            cash_amount = cur_obj.compute(cr, uid, journal.currency.id, invoice.company_id.currency_id.id, cash_amount, context=ctx)
            context.update({'cash_amount':cash_amount})
            
        if invoice.company_id.currency_id.id<>invoice.currency_id.id:    
            cash_amount = cur_obj.compute(cr, uid, journal.currency.id, invoice.company_id.currency_id.id, cash_amount, context=ctx)
            context.update({'cash_amount':cash_amount})
            
        acc_id = journal.default_credit_account_id and journal.default_credit_account_id.id
        if not acc_id:
            raise osv.except_osv(_('Error !'), _('Your journal must have a default credit and debit account.'))

        self.pool.get('account.invoice').pay_and_reconcile(cr, uid, [context['id']],
                amount, acc_id, data['period_id'], data['journal_id'], writeoff_account_id,
                data['period_id'], writeoff_journal_id, context, data['name'])
        return {}
    
    def fields_view_get(self, cr, uid, view_id=None, view_type='form', context=None, toolbar=False):
        
        res = super(account_invoice_pay,self).fields_view_get(cr, uid, view_id, view_type, context, toolbar)
        fields=res.get('fields',{})
        word = "Supplier Cash Discount Account"
        type = context.get('type',False)
        eview = etree.fromstring(res['arch'])
        res['arch'] = etree.tostring(eview)
        if type == 'in_invoice':
            for f in res['fields']:
                if 'Customers Cash Discount Account' in res['fields'][f]['string']:
                    res['fields'][f]['string'] = res['fields'][f]['string'].replace('Customers Cash Discount Account',word)
        return res
    
    def _cancel_line(self, cr, uid, ids, context=None):
         move_obj = self.pool.get('account.move')
         move_line_obj = self.pool.get('account.move.line')
         data =  self.read(cr, uid, ids,context=context)[0]

         if data['tax_move_ids']:
            move_line = data['tax_move_ids']
            for move_line_id in move_line:
                move_id = move_line_obj.browse(cr, uid,move_line_id).move_id.id
                move_line_obj.unlink(cr, uid,[move_line_id])
                move_obj.unlink(cr, uid,[move_id])
                
         if data['discount_move_ids']: 
            discount_move_line = data['discount_move_ids']
            for move_line_id in discount_move_line:
                move_id = move_line_obj.browse(cr, uid,move_line_id).move_id.id
                move_line_obj.unlink(cr, uid,[move_line_id])
                move_obj.unlink(cr, uid,[move_id])

         return {}

account_invoice_pay()

class account_message(osv.osv_memory):
    _name = "account.message"
    def _check(self, cr, uid, ids, context=None):
        data =  self.read(cr, uid, ids,context=context)[0]
        result = self.pool.get('account.invoice.pay').wo_check(cr, uid, ids, context=context)
        return result

account_message()    

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4: